/**
 * @file aisl/instance.h
 *
 * Copyright (c) 2017-2019 by Löwenware Ltd.
 *
 * Project homepage: https://lowenware.com/aisl/
 *
 */

#ifndef AISL_INSTANCE_H_60576F41_454C_4189_B91A_F40501132230
#define AISL_INSTANCE_H_60576F41_454C_4189_B91A_F40501132230

#include <stdint.h>
#include <stdarg.h>
#include <aisl/types.h>
#include <aisl/config.h>



/**
 * @brief Allocates new AISL instance.
 *
 * @param config a pointer to #aisl_config structure.
 * @return an #aisl_t instance pointer.
 */
aisl_t
aisl_new( aisl_config_t config );


/**
 * @brief Frees previously allocated pointer of AISL instance.
 * @param instance a pointer to #aisl_t instance.
 */
void
aisl_free( aisl_t instance );


/**
 * @brief Allocates and registers HTTP server instance.
 * @param instance a pointer to #aisl_t instance.
 * @param address a null-terminated ip or hostname string.
 * @param port a number of port to listen.
 * @return a pointer representing HTTP server.
 */
aisl_server_t
aisl_listen( aisl_t instance, const char * address, uint16_t port );


#ifndef AISL_WITHOUT_SSL

/** 
 * @brief Sets pair of SSL certificate and key for domain name.
 * @param instance a pointer to #aisl_t instance.
 * @param domain a null-terminated string with domain name.
 * @param key_file a null-terminated string with path to private SSL key file.
 * @param crt_file a null-terminated string with path to SSL certificate file.
 * @return #aisl_status_t code.
 */
aisl_status_t
aisl_set_ssl( aisl_t           instance,
              const char     * domain,
              const char     * key_file,
              const char     * crt_file );

#endif


/** 
 * @brief Registers a callback for event and its source.
 * If source is NULL, then callback will be executed for all events of specified
 * type.
 *
 * Typical sources are:
 *   - #aisl_server_t,
 *   - #aisl_client_t,
 *   - #aisl_stream_t;
 *
 * @param instance a pointer to #aisl_t instance.
 * @param event a code of event.
 * @param callback a pointer to function that will be triggered on event.
 */
void
aisl_set_callback( aisl_t             instance,
                   aisl_event_t       event,
                   aisl_callback_t    callback );


/** 
 * @brief A core function doing all the library routines.
 * Designed to be called inside application main loop
 * @param instance a pointer to #aisl_t instance.
 * @return #aisl_status_t code.
 */
aisl_status_t
aisl_run_cycle( aisl_t instance );


/** 
 * @brief Function to sleep CPU if nothing to do.
 * Calls select on all the opened sockets inside.
 * @param instance a pointer to #aisl_t instance.
 * @param usec a number of miliseconds to wait for any data on sockets.
 * @return #aisl_status_t code.
 */
aisl_status_t
aisl_sleep( aisl_t instance, uint32_t usec );


/** 
 * @brief Get last error message.
 * @param instance a pointer to #aisl_t instance.
 * @return a null-terminated string with error message.
 */
const char *
aisl_get_error( aisl_t instance );

#endif /* !AISL_INSTANCE_H */
