/*
 * src/instance.c
 *
 * Copyright (C) 2019 Ilja Kartašov <ik@lowenware.com>
 *
 * Project homepage: https://lowenware.com/aisl/
 *
 */

#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/select.h>
#include <sys/time.h>

#ifndef AISL_WITHOUT_SSL
#include <openssl/err.h>
#endif

#include "debug.h"
#include "str-utils.h"
#include "buffer.h"

#include "client.h"
#include "server.h"
#include "ssl.h"
//#include "globals.h"
#include "stream.h"
#include "instance.h"


#ifndef AISL_WITHOUT_SSL
static uint32_t m_instances = 0;
#endif


/* Initialization functions */

__attribute__ ((visibility ("default") ))
aisl_t
aisl_new( aisl_config_t config )
{
  aisl_t instance;

  #ifndef AISL_WITHOUT_SSL
  if ((m_instances++) == 0)
  {
    SSL_load_error_strings();
    OpenSSL_add_ssl_algorithms();
  }
  #endif

  if ( !(instance = calloc(1, sizeof(struct aisl))) )
    goto finally;

  if ( list_init(&instance->server_spool, config->server_spool_size) == -1 )
    goto release;

  if ( list_init(&instance->client_spool, config->client_spool_size) == -1 )
    goto release;

  #ifndef AISL_WITHOUT_SSL
  if ( list_init(&instance->ssl_spool, config->ssl_spool_size) == -1 )
    goto release;
  #endif

  instance->accept_limit = config->client_accept_limit;
  instance->silence_timeout = config->client_silence_timeout;

  goto finally;

release:
  aisl_free(instance);
  instance = NULL;

finally:
  return instance;
}


__attribute__ ((visibility ("default") ))
void
aisl_free( aisl_t instance )
{
  list_release(&instance->client_spool, (list_destructor_t) aisl_client_free );

  list_release(&instance->server_spool, (list_destructor_t)  aisl_server_free );

  #ifndef AISL_WITHOUT_SSL
  list_release(&instance->ssl_spool, (list_destructor_t) aisl_ssl_free );
  #endif

  if (instance->last_error)
    free(instance->last_error);

  free(instance);

  #ifndef AISL_WITHOUT_SSL
  if ((--m_instances) == 0)
  {
    EVP_cleanup();
  }
  #endif
}


__attribute__ ((visibility ("default") ))
aisl_server_t
aisl_listen( aisl_t instance, const char * address, uint16_t port )
{
  aisl_server_t result;

  if ( (result = aisl_server_new(instance, address, port)) != NULL )
  {
    if (list_append(&instance->server_spool, result) == -1)
    {
      aisl_server_free(result);
    }
  }

  return result;
}


#ifndef AISL_WITHOUT_SSL


static int
aisl_on_get_ssl( SSL * ssl, int * ptr, void * instance )
{
  const char * server_name = SSL_get_servername(ssl, TLSEXT_NAMETYPE_host_name);

  SSL_CTX * ctx = aisl_get_ssl_ctx( (aisl_t) instance, server_name );

  if (ctx)
  {
    SSL_set_SSL_CTX(ssl, ctx);
  }

  (void)ptr;

  return SSL_TLSEXT_ERR_OK;
}


static aisl_status_t
aisl_set_ssl_ctx(aisl_t instance, aisl_ssl_t ssl)
{
  const SSL_METHOD * method;
  SSL_CTX          * ctx;

  method = SSLv23_server_method();

  if ( (ctx = SSL_CTX_new(method)) != NULL )
  {
    SSL_CTX_set_ecdh_auto(ctx, 1);

    SSL_CTX_set_tlsext_servername_callback( ctx, aisl_on_get_ssl );
    SSL_CTX_set_tlsext_servername_arg(      ctx, (void *) instance );

    if (!(SSL_CTX_use_certificate_file(ctx, ssl->crt_file, SSL_FILETYPE_PEM)>0))
      goto except;

    if (!(SSL_CTX_use_PrivateKey_file(ctx, ssl->key_file, SSL_FILETYPE_PEM)>0))
      goto except;

    ssl->ctx = ctx;

    return AISL_SUCCESS;
  }

except:
  SSL_CTX_free(ctx);
  aisl_set_error( instance, ERR_error_string(ERR_get_error(),NULL) );
  return AISL_SYSCALL_ERROR;
}


__attribute__ ((visibility ("default") ))
aisl_status_t
aisl_set_ssl( aisl_t instance, const char * host,
                               const char * key_file,
                               const char * crt_file )
{
  SSL_CTX * ssl_ctx = NULL;
  aisl_ssl_t ssl;
  size_t i;

  /* lookup for existing contexts */
  for (i=0; i<instance->ssl_spool.count; i++)
  {
    ssl = LIST_INDEX(instance->ssl_spool, i);
    if (ssl->key_file && strcmp(ssl->key_file, key_file)==0 &&
        ssl->crt_file && strcmp(ssl->crt_file, crt_file)==0 )
    {
      ssl_ctx = ssl->ctx;
      break;
    }
  }

  if ((ssl = aisl_ssl_new(host, key_file, crt_file, ssl_ctx)) != NULL)
  {
    if (!ssl_ctx || (aisl_set_ssl_ctx(instance, ssl) == AISL_SUCCESS))
    {
      if (list_append(&instance->ssl_spool, ssl) != -1)
        return AISL_SUCCESS;
    }

    aisl_ssl_free(ssl);
  }

  return AISL_MALLOC_ERROR;
}


SSL_CTX *
aisl_get_ssl_ctx( aisl_t instance, const char * host )
{
  size_t i;
  aisl_ssl_t ssl;

  if (host)
  {
    for (i=0; i<instance->ssl_spool.count; i++)
    {
      ssl = LIST_INDEX(instance->ssl_spool, i);
      if (strcmp(ssl->host, host) != 0)
        continue;

      return ssl->ctx;
    }
  }

  return NULL;
}


#endif

__attribute__ ((visibility ("default") ))
void
aisl_set_callback( aisl_t           instance,
                   aisl_event_t     event,
                   aisl_callback_t  callback  )
{
  switch( event )
  {
    case AISL_EVENT_SERVER_OPEN:
      instance->on_server_open = (aisl_on_server_open_t) callback;
      break;

    case AISL_EVENT_SERVER_ERROR:
      instance->on_server_error = (aisl_on_server_error_t) callback;
      break;

    case AISL_EVENT_CLIENT_CONNECT:
      instance->on_client_connect = (aisl_on_client_connect_t) callback;
      break;

    case AISL_EVENT_CLIENT_DISCONNECT:
      instance->on_client_disconnect = (aisl_on_client_disconnect_t) callback;
      break;

    case AISL_EVENT_CLIENT_TIMEOUT:
      instance->on_client_timeout = (aisl_on_client_timeout_t) callback;
      break;

    case AISL_EVENT_STREAM_OPEN:
      instance->on_stream_open = (aisl_on_stream_open_t) callback;
      break;

    case AISL_EVENT_STREAM_HEADER:
      instance->on_stream_header = (aisl_on_stream_header_t) callback;
      break;

    case AISL_EVENT_STREAM_INPUT:
      instance->on_stream_input = (aisl_on_stream_input_t) callback;
      break;

    case AISL_EVENT_STREAM_REQUEST:
      instance->on_stream_request = (aisl_on_stream_request_t) callback;
      break;

    case AISL_EVENT_STREAM_OUTPUT:
      instance->on_stream_output = (aisl_on_stream_output_t) callback;
      break;

    case AISL_EVENT_STREAM_CLOSE:
      instance->on_stream_close = (aisl_on_stream_close_t) callback;
      break;

    case AISL_EVENT_STREAM_ERROR:
      instance->on_stream_error = (aisl_on_stream_error_t) callback;
      break;
  }
}


void
aisl_raise( aisl_t instance, void * source, aisl_event_t event, ... )
{
  DPRINTF("event %s", aisl_event_to_string(event));

  va_list vl;

  va_start(vl, event);
  switch( event )
  {
    case AISL_EVENT_SERVER_OPEN:
      if (instance->on_server_open)
        instance->on_server_open((aisl_server_t)source);
      break;

    case AISL_EVENT_SERVER_ERROR:
      if (instance->on_server_error)
        instance->on_server_error((aisl_server_t)source);
      break;

    case AISL_EVENT_CLIENT_CONNECT:
      if (instance->on_client_connect)
        instance->on_client_connect((aisl_client_t)source);
      break;

    case AISL_EVENT_CLIENT_DISCONNECT:
      if (instance->on_client_disconnect)
        instance->on_client_disconnect((aisl_client_t)source);
      break;

    case AISL_EVENT_CLIENT_TIMEOUT:
      if (instance->on_client_timeout)
        instance->on_client_timeout((aisl_client_t)source);
      break;

    case AISL_EVENT_STREAM_OPEN:
      if (instance->on_stream_open)
      {
        aisl_http_method_t method = va_arg(vl, aisl_http_method_t);
        const char * path = va_arg(vl, const char *);
        const char * query = va_arg(vl, const char *);
        instance->on_stream_open((aisl_stream_t)source, method, path, query);
      }
      break;

    case AISL_EVENT_STREAM_HEADER:
      if (instance->on_stream_header)
      {
        const char * key = va_arg(vl, const char *);
        const char * val = va_arg(vl, const char *);
        instance->on_stream_header((aisl_stream_t)source, key, val);
      }
      break;

    case AISL_EVENT_STREAM_INPUT:
      if (instance->on_stream_input)
      {
        const char * data = va_arg(vl, const char *);
        int32_t      size = va_arg(vl, int32_t);
        instance->on_stream_input((aisl_stream_t)source, data, size);
      }
      break;

    case AISL_EVENT_STREAM_REQUEST:
      if (instance->on_stream_request)
        instance->on_stream_request((aisl_stream_t)source);
      break;

    case AISL_EVENT_STREAM_OUTPUT:
      if (instance->on_stream_output)
        instance->on_stream_output((aisl_stream_t)source);
      break;

    case AISL_EVENT_STREAM_CLOSE:
      if (instance->on_stream_close)
        instance->on_stream_close((aisl_stream_t)source);
      break;

    case AISL_EVENT_STREAM_ERROR:
      if (instance->on_stream_error)
        instance->on_stream_error((aisl_stream_t)source);
      break;
  }
  va_end(vl);
}


__attribute__ ((visibility ("default") ))
aisl_status_t
aisl_run_cycle( aisl_t instance )
{
  aisl_status_t result = AISL_IDLE;

  int32_t i = instance->iterator,
          max = instance->server_spool.count + instance->client_spool.count;

  while ( result == AISL_IDLE && max != 0)
  {
    aisl_client_t cli = NULL;

    if (i < instance->server_spool.count)
    {
      if (instance->client_spool.count < instance->accept_limit)
      {
        aisl_server_t srv;
        /* SSL_CTX * ssl_ctx; */

        srv = (aisl_server_t) LIST_INDEX(instance->server_spool, i);

        /*
        ssl_ctx = (aisl_server_get_ssl(srv)) ?
                    aisl_get_ssl_ctx(instance, NULL) :
                    NULL;
        */
        result = aisl_server_touch(srv, &cli);

        if (cli)
        {
          DPRINTF("Accepted %p", (void*)cli);
          if ( list_append(&instance->client_spool, cli) != -1 )
          {
            aisl_raise(instance, cli, AISL_EVENT_CLIENT_CONNECT);
          }
          else
          {
            aisl_client_free(cli);
            /* todo: save error message here */
          }
        }
      }
      else
      {
        i = instance->server_spool.count;
        continue;
      }
    }
    else
    {
      int32_t j = i - instance->server_spool.count;

      if (j < instance->client_spool.count)
      {
        aisl_client_t c = LIST_INDEX(instance->client_spool, j);

        result = aisl_client_touch(c);

        if (aisl_client_is_timed_out( c, instance->silence_timeout ) )
          aisl_raise( instance, c, AISL_EVENT_CLIENT_TIMEOUT );

        if ( !aisl_client_is_online(c) )
        {
          aisl_client_free( c );
          list_remove_index(&instance->client_spool, j);
        }
      }
      else
      {
        i = 0;
        continue;
      }
    }

    max--;
    i++;
  }

  instance->iterator = i;

  return result;
}


__attribute__ ((visibility ("default") ))
const char *
aisl_get_error( aisl_t instance )
{
  return instance->last_error;
}


void
aisl_set_error( aisl_t instance, const char * err_msg )
{
  if (instance->last_error)
    free(instance->last_error);

  instance->last_error = str_copy(err_msg);
}


__attribute__ ((visibility ("default") ))
aisl_status_t
aisl_sleep( aisl_t instance, uint32_t usec )
{
  int maxfd=0,
      sd;
  size_t i;
  struct timeval timeout = {0,usec};

  memset(&timeout, 0, sizeof(struct timeval));
  timeout.tv_usec = usec;

  fd_set  fs;
  FD_ZERO (&fs);

  for (i=0; i<instance->server_spool.count; i++)
  {
    aisl_server_t s = LIST_INDEX(instance->server_spool, i);

    sd = aisl_server_get_socket(s);

    if (sd != -1)
    {
      FD_SET(sd, &fs);
      if (sd > maxfd) maxfd = sd;
    }
  }


  for (i=0; i<instance->client_spool.count; i++)
  {
    aisl_client_t c = LIST_INDEX(instance->client_spool, i);
    sd = aisl_client_get_socket(c);
    if (sd != -1)
    {
      FD_SET(sd, &fs);
      if (sd > maxfd) maxfd = sd;
    }
  }

  switch ( select(maxfd+1, &fs, NULL, NULL, &timeout) )
  {
    case -1:
      return AISL_SYSCALL_ERROR;

    case 0:
      return AISL_IDLE;

    default:
      return AISL_SUCCESS;
  }
}

